/* User, channel and server routines.

   OperStats - Copyright (c) 2000-2001 Darcy Grexton
   Contact: cygnus@uniserve.ca, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/operstats.h"

Channel *chanlist[1024];
User *userlist[1024];
Clone *clonelist[1024];

/* Add a new user into memory. */
static User *new_user (const char *nick)
{
    User *user, **list;

    user = scalloc (sizeof (User), 1);
    if (!nick)
	nick = "";
    strscpy (user->nick, nick, NICKLEN);
    list = &userlist[NHASH(user->nick)];
    user->next = *list;
    if (*list)
	(*list)->prev = user;
    *list = user;
    usercnt++;
    userstat++;
    dailyhitcnt++;
    if (userstat > maxusercnt)
    {
	maxusercnt = userstat;
	maxusertime = time (NULL);
    }
    if (userstat > dailyuserstat)
    {
	dailyuserstat = userstat;
	dailyusertime = time (NULL);
    }
    if (dailyhitcnt > maxhitcnt)
	maxhitcnt = dailyhitcnt;
    return user;
}

/* Find a user by nick. Return NULL if user could not be found. */
User *finduser (const char *nick)
{
    User *user;

    user = userlist[NHASH(nick)];
    while (user && stricmp (user->nick, nick))
	user = user->next;
    return user;
}

/* Add a new clone to the list */ 
static Clone *new_clone (const char *host)
{
    Clone *clone, **list;

    clone = scalloc (sizeof (Clone), 1);
    if (!host)
	host = "";
    strscpy (clone->host, host, HOSTLEN);
    list = &clonelist[CLHASH(clone->host)];
    clone->next = *list;
    if (*list)
	(*list)->prev = clone;
    *list = clone;
    return clone;
}

/* Find a Clone. Return NULL if clone could not be found. */
Clone *findclone (const char *host)
{
    Clone *clone;

    clone = clonelist[CLHASH(host)];
    while (clone && stricmp (clone->host, host))
	clone = clone->next;
    return clone;
}

/* Add a clone, creating the structure if needed. */
static void add_clone (User *u)
{
    Clone *clone;
    Trigger *trigger;
    int limit;

    if (!(clone = findclone (u->host)))
	clone = new_clone (u->host);

    clone->cnt++;

    /* Check for a trigger for this host */
    if ((trigger = findtrig (u->user, u->host)))
	limit = trigger->limit;
    else
	limit = maxclones;

    if (clone->cnt > limit)
    {
	/* If they've broke the max clone limit, akill them */
	if (maxclonekills && clone->kills >= maxclonekills)
	{
	    AutoKill *akill, **list;
	    char *mask;

	    mask = smalloc (strlen (u->user) + strlen (u->host) + 2);
	    snprintf (mask, strlen (u->user) + strlen (u->host) + 2, "*@%s",
		u->host);

	    /* Ignore it if we already have it. (which we shouldn't) */
	    if (is_akilled (mask))
		return;

	    akill = scalloc (sizeof (AutoKill), 1);
	    akill->mask = sstrdup (mask);
	    akill->reason = sstrdup (OS_MAX_CLONES);
	    akill->setter = sstrdup (me.name);
	    akill->realname = 0;
	    akill->set = time (NULL);
	    akill->expires = def_akill_time;
	    list = &akilllist[AHASH(akill->mask)];
	    akill->next = *list;
	    if (*list)
		(*list)->prev = akill;
	    *list = akill;

#ifdef IRCUP9
	    /* We use 999999999 because we'll expire it ourselves. */
	    send_cmd (me.name, "GLINE * +*@%s 999999999 :%s",
		u->host, akill->reason);
#elif defined (BAHAMUT)
	    send_cmd (me.name, "AKILL %s * 0 %s %ld :%s", u->host,
		s_OperServ, time (NULL), akill->reason);
#else
	    send_cmd (me.name, "%s %s * :%s", me.token ? "V" : "AKILL",
		u->host, akill->reason);
#endif

	    free (mask);
	}
	kill_user (s_OperServ, u->nick, OS_MAX_CLONES);

	if (maxclonekills)
	    clone->kills++;
    }
}

/* Remove a clone from a host, deleting it as needed */
static void del_clone (User *u)
{
    Clone *clone;

    if (!(clone = findclone (u->host)))
	return;

    clone->cnt--;

    if (clone->cnt <= 0)
    {
	if (clone->prev)
	    clone->prev->next = clone->next;
	else
	    clonelist[CLHASH(clone->host)] = clone->next;
	if (clone->next)
	    clone->next->prev = clone->prev;

	free (clone);
    }
}

/* Return the Channel structure corresponding to the named channel, or
   NULL if the channel was not found.
 */
Channel *findchan (const char *chan)
{
    Channel *c;

    c = chanlist[CHASH(chan)];
    while (c)
    {
	if (!stricmp (c->name, chan))
	    return c;
	c = c->next;
    }
    return NULL;
}

/* Change a user's nick */
static void change_user_nick (User *user, const char *nick)
{
    User **list;

    if (user->prev)
	user->prev->next = user->next;
    else
	userlist[NHASH(user->nick)] = user->next;
    if (user->next)
	user->next->prev = user->prev;
    user->nick[1] = 0;
    strscpy (user->nick, nick, NICKLEN);
    list = &userlist[NHASH(user->nick)];
    user->next = *list;
    user->prev = NULL;
    if (*list)
	(*list)->prev = user;
    *list = user;
}

/* Handle a server NICK command.
	av[0] = nick
	If a new user:
	   av[1] = hop count
	   av[2] = signon time
	   DreamForge:
		av[3] = username
		av[4] = hostname
		av[5] = server
		av[6] = servicestamp
		av[7] = realname
	   IRCu:
		av[3] = username
		av[4] = hostname
		av[5] = server
		av[6] = realname
	   Bahamut:
		av[3] = modes
		av[4] = username
		av[5] = hostname
		av[6] = server
		av[7] = servicestamp
		av[8] = realname
	   Unreal:
		av[3] = username
		av[4] = hostname
		av[5] = server
		av[6] = servicestamp
		av[7] = modes
		av[8] = vhost
		av[9] = realname
	Else:
	  av[1] = time of change
 */
void do_nick (const char *source, int ac, char **av)
{
    User *user;
    char *tmp;

    if (!*source)
    {
	MSG *msg;
	TLD *tld;

	/* This is a new user; create a User structure for it. */
	if (debuglevel > 1)
	    debug ("do_nick(): %s", av[0]);

	/* Check if this user matches any AKills */
#ifdef BAHAMUT
	if (check_akill (av[0], av[4], av[5], av[8]))
	    return;
#elif defined (IRCUP9)
	if (check_akill (av[0], av[3], av[4], av[6]))
	    return;
#else
	if (check_akill (av[0], av[3], av[4], av[7]))
	    return;
#endif

	/* Allocate User structure and fill it in. */
	user = new_user (av[0]);
#ifdef BAHAMUT
	user->user	= sstrdup (av[4]);
	user->host	= sstrdup (av[5]);
	user->server	= sstrdup (av[6]);
	user->real	= sstrdup (av[8]);

	av[1] = av[3];
	do_umode (av[0], 2, av);
#elif defined (IRCUP9)
	user->user	= sstrdup (av[3]);
	user->host	= sstrdup (av[4]);
	user->server	= sstrdup (av[5]);
	user->real	= sstrdup (av[6]);
#elif defined (UNREAL3)
	user->user	= sstrdup (av[3]);
	user->host	= sstrdup (av[4]);
	user->server	= sstrdup (av[5]);
	user->real	= sstrdup (av[9]);

	av[1] = av[7];
	do_umode (av[0], 2, av);
#else
	user->user	= sstrdup (av[3]);
	user->host	= sstrdup (av[4]);
	user->server	= sstrdup (av[5]);
	user->real	= sstrdup (av[7]);
#endif /* BAHAMUT */

	/* Show them the LogonMSGs */
	if (globalnoticer_on == TRUE)
	    for (msg = msglist; msg; msg = msg->next)
		notice (s_GlobalNoticer, av[0], msg->text);

	/* Increase the count for their TLD */
	tmp = strrchr (user->host, '.');
	tmp = strtok (tmp, ".");
	if ((tld = findtld (tmp)))
	    ++tld->cnt;
	else
	    ++tldnores;

	/* Clone checking stuff */
	if (maxclones)
	    add_clone (user);
    }
    else
    {
	/* An old user changing nicks. */
	user = finduser (source);

	if (!user)
	{
	    if (debuglevel > 1)
		debug ("do_nick(): Got nickchange for nonexistant user %s->%s",
		    source, av[0]);
	    return;
	}

	if (debuglevel > 1)
	    debug ("do_nick(): %s changed nick to %s", source, av[0]);

	/* If this is a nickname we were waiting for, reclaim it. */
	if (os.wantnick)
	    if (!stricmp (source, os.nick))
	    {
		send_cmd (s_OperServ, "%s %s", me.token ? "&" : "NICK",
		    os.nick);
		strscpy (s_OperServ, os.nick, sizeof (s_OperServ));
		os.wantnick = 0;
	    }

	if (ss.wantnick)
	    if (!stricmp (source, ss.nick))
	    {
		send_cmd (s_StatServ, "%s %s", me.token ? "&" : "NICK",
		    ss.nick);
		strscpy (s_StatServ, ss.nick, sizeof (s_StatServ));
		ss.wantnick = 0;
	    }

	if (gn.wantnick)
	    if (!stricmp (source, gn.nick))
	    {
		send_cmd (s_GlobalNoticer, "%s %s", me.token ? "&" : "NICK",
		    gn.nick);
		strscpy (s_GlobalNoticer, gn.nick, sizeof (s_GlobalNoticer));
		gn.wantnick = 0;
	    }

	change_user_nick (user, av[0]);
    }
}

/* Handle a user modechange - This currently only cares about
   umode o, since we don't need any others.
 */
void do_umode (const char *source, int ac, char **av)
{
    User *user;
    char *s;
    int add = 1;

    if (stricmp (source, av[0]))
	return;

    user = finduser (source);

    if (!user)
    {
	if (debuglevel > 1)
	    debug ("MODE for nonexistant user %s: %s", source, av[1]);
	return;
    }

    s = av[1]; 

    while (*s)
    {
	switch (*s++)
	{
	    case '+':
		add = 1;
		break;
	    case '-':
		add = 0;
		break;
	    case 'o':
		add ? (user->mode |= UMODE_o) : (user->mode &= ~UMODE_o);
		if (add == 0 && (user->flags & UF_SRA))
		    user->flags &= ~UF_SRA;
		break;
	}
    }
}

/* Handle a channel modechange - Only cares about cmode s since it's
   the only one we need.
 */
void do_cmode (const char *source, int ac, char **av)
{
    Channel *chan;
    char *s;
    int add = 1;
    char *modestr = av[1];

    chan = findchan (av[0]);

    if (!chan)
    {
	if (debuglevel > 1)
	    debug ("MODE for nonexistant channel %s: %s", av[0], av[1]);
	return;
    }

    s = modestr;
    ac -= 2;
    av += 2;

    while (*s)
    {
	switch (*s++)
	{
	    case '+':
		add = 1;
		break;
	    case '-':
		add = 0;
		break;
	    case 's':
		add ? (chan->mode |= CMODE_s) : (chan->mode &= ~CMODE_s);
		break;
	}
    }
}

/* Remove and free a User structure. */
void delete_user (User * user)
{
    struct u_chanlist *c, *c2;
    TLD *tld;
    char *tmp;

    if (debuglevel > 1)
	debug ("delete_user(): %s", user->nick);

    usercnt--;
    userstat--;

    /* Decrease the count for their TLD */
    tmp = strrchr (user->host, '.');
    tmp = strtok (tmp, ".");
    if ((tld = findtld (tmp)))
	--tld->cnt;
    else
	--tldnores;

    /* Remove a clone from their record */
    del_clone (user);

    free (user->user);
    free (user->host);
    free (user->real);
    free (user->server);

    c = user->chans;

    while (c)
    {
	c2 = c->next;
	chan_deluser (user, c->chan);
	free (c);
	c = c2;  
    }

    if (user->prev)
	user->prev->next = user->next;
    else
	userlist[NHASH(user->nick)] = user->next;
    if (user->next)
	user->next->prev = user->prev;

    free (user);
}

/* Handle a JOIN command.
      av[0] = channels to join
 */
void do_join (const char *source, int ac, char **av)
{
    User *user;
    char *s, *t;
    struct u_chanlist *c, *c2;

    user = finduser (source);

    if (!user)
    {
	if (debuglevel > 1)
	    debug ("do_join(): Got JOIN to %s from nonexistant user %s",
		av[0], source);
	return;
    }

    t = av[0];

    while (*(s = t))
    {
	t = s + strcspn (s, ",");
	if (*t)
	    *t++ = 0;
        if (*s == '0')
        {
	    c = user->chans;
	    while (c)
	    {
		c2 = c->next;
		chan_deluser (user, c->chan);
		if (debuglevel > 1)
		    debug ("do_join(): %s parted all channels");
		free (c);
		c = c2;
	    }
	    user->chans = NULL;
	    continue;
	}
	if (debuglevel > 1)
	    debug ("do_join(): %s joined %s", source, s);
	chan_adduser (user, s);
	c = smalloc (sizeof (*c));
	c->next = user->chans; 
	c->prev = NULL;
	if (user->chans)
	    user->chans->prev = c;
	user->chans = c;
	c->chan = findchan (s);
    }
}

/* Handle a PART command */
void do_part (const char *source, int ac, char **av)
{
    User *user;
    char *s, *t;
    struct u_chanlist *c;
 
    user = finduser (source);

    if (!user)  
	return;

    t = av[0];

    while (*(s = t))
    {
	t = s + strcspn (s, ",");
	if (*t) 
	    *t++ = 0;
	if (debuglevel > 1)
	    debug ("do_part(): %s parted %s", source, s);
	for (c = user->chans; c && stricmp (s, c->chan->name) != 0; c = c->next)
	    ;  
	if (c)
	{
	    chan_deluser (user, c->chan);
	    if (c->next)
		c->next->prev = c->prev;
	    if (c->prev)
		c->prev->next = c->next;
	    else
		user->chans = c->next;
	    free (c);
	}
    }
}

/* Handle a KICK command */
void do_kick (const char *source, int ac, char **av)
{
    User *user;
    char *s, *t;
    struct u_chanlist *c;   
        
    if (is_one_of_mine (av[1]))
    {
	/* Rejoin our channel if kicked */
	if (!stricmp (av[0], joinchan))
	{
#ifdef BAHAMUT
	    send_cmd (av[1], "SJOIN %ld %ld %s + :%s", time (NULL), time (NULL),
		joinchan, av[1]);
#else
	    send_cmd (av[1], "JOIN %s", joinchan);
#endif
	    send_cmd (av[1], "MODE %s +o %s", joinchan, av[1]);
	}
	return;
    }

    t = av[1];

    while (*(s = t))
    {
	t = s + strcspn (s, ",");
	if (*t)
	    *t++ = 0;
	user = finduser (s);
	if (!user)
	{
	    if (debuglevel > 1)
		debug ("do_kick(): Got KICK for nonexistant user %s from %s",
		    user->nick, av[0]);
	    continue;
	}
	if (debuglevel > 1)
	    debug ("do_kick(): %s was kicked from %s", s, av[0]);
	for (c = user->chans; c && stricmp (av[0], c->chan->name) != 0; c = c->next)
	    ;
	if (c) 
	{
	    chan_deluser (user, c->chan);
	    if (c->next)
		c->next->prev = c->prev;
	    if (c->prev)
		c->prev->next = c->next;
	    else
		user->chans = c->next;
	    free (c);
	}
    }
    if (!source)
	return;
}

/* Handle a QUIT command */
void do_quit (const char *source, int ac, char **av)
{    
    User *user;

    user = finduser (source);

    if (!user)
    {
	if (debuglevel > 1)
	    debug ("do_quit(): Got QUIT for nonexistant user %s", user->nick);
	return;
    }

    if (debuglevel > 1)
	debug ("do_quit(): %s QUIT", source);

    /* If this is a nickname we were waiting for, reclaim it. */
    if (os.wantnick)
	if (!stricmp (source, os.nick))
	{
	    send_cmd (s_OperServ, "%s %s", me.token ? "&" : "NICK", os.nick);
	    strscpy (s_OperServ, os.nick, sizeof (s_OperServ));
	    os.wantnick = 0;
	}

    if (ss.wantnick)
	if (!stricmp (source, ss.nick))
	{
	    send_cmd (s_StatServ, "%s %s", me.token ? "&" : "NICK", ss.nick);
	    strscpy (s_StatServ, ss.nick, sizeof (s_StatServ));
	    ss.wantnick = 0;
	}

    if (gn.wantnick)
	if (!stricmp (source, gn.nick))
	{
	    send_cmd (s_GlobalNoticer, "%s %s", me.token ? "&" : "NICK",
		gn.nick);
	    strscpy (s_GlobalNoticer, gn.nick, sizeof (s_GlobalNoticer));
	    gn.wantnick = 0;
	}

    delete_user (user);
}

/* Add a user to a channel, creating the channel as necessary. */
void chan_adduser (User * user, const char *chan)
{
    Channel *c = findchan (chan);
    Channel **list;
    int newchan = !c;
    struct c_userlist *u;

    if (newchan)
    {
	/* Allocate pre-cleared memory */
	c = scalloc (sizeof (Channel), 1);
	strscpy (c->name, chan, sizeof (c->name));
	list = &chanlist[CHASH(c->name)];
	c->next = *list;
	if (*list)
	    (*list)->prev = c;
	*list = c;
	++chancnt;
	++chanstat;
	if (chanstat > maxchancnt)
	{
	    maxchancnt = chanstat;
	    maxchantime = time (NULL);
	}
	if (chanstat > dailychanstat)
	{
	    dailychanstat = chanstat;
	    dailychantime = time (NULL);
	}
    }

    u = smalloc (sizeof (struct c_userlist));
    u->next = c->users;
    u->prev = NULL;

    if (c->users)
	c->users->prev = u;

    c->users = u;
    u->user = user;
}

/* Remove a user from a channel, deleting the channel as necessary. */
void chan_deluser (User * user, Channel * c)
{
    struct c_userlist *u;

    if (debuglevel > 1)
	debug ("chan_deluser(): %s %s", user->nick, c->name);

    for (u = c->users; u && u->user != user; u = u->next)
	;

    if (!u)
    {
	if (debuglevel > 1)
	    debug ("chan_deluser(): %s not in %s!", user->nick, c->name);
    }
    else
    {
	if (u->next)
	    u->next->prev = u->prev;
	if (u->prev)
	    u->prev->next = u->next;
	else
	    c->users = u->next;
	free (u);
    }

    if (!c->users)
    {
	if (debuglevel > 1)
	    debug ("chan_deluser(): Deleting empty channel %s", c->name);
	--chancnt;
	--chanstat;
	if (c->next)
	    c->next->prev = c->prev;
	if (c->prev)
	    c->prev->next = c->next;
	else
	    chanlist[CHASH(c->name)] = c->next;
	free (c);
    }
}

/* Change a channel's stored topic. */
void do_topic (const char *source, int ac, char **av)
{
    Channel *c = findchan (av[0]);

    if (!c)
	return;

    if (c->topic)
    {
	free (c->topic);
	c->topic = NULL;
    }

    if (ac > 3 && *av[3])
	c->topic = sstrdup (av[3]);
}

/* Handle a KILL command.
      av[0] = nick being killed
      av[1] = reason
 */
void do_kill (const char *source, int ac, char **av)
{
    User *user;

    user = finduser (av[0]);

    if (!user)
    {
	if (debuglevel > 1)
	    debug ("do_kill(): Got kill for nonexistant user %s", av[0]);
	return;
    }

    if (debuglevel > 1)
	debug ("do_kill(): %s Killed", av[0]);
    delete_user (user);
}

/* Return channel statistics - This is mainly so StatServ's counts
   match lusers, which doesn't include +s channels. This returns one
   of two things: The total amount of channels that aren't +s, or the
   total amount that are.
 */
int countchans (int secret)
{
    Channel *c;
    int i, total = 0;

    for (i = 0; i < 1024; i++)
    {
	for (c = chanlist[i]; c; c = c->next)
	{
	    if (secret && (c->mode & CMODE_s))
		++total;
	    else if (!secret && !(c->mode & CMODE_s))
		++total;
	}
    }
    return total;
}

/* Remove a user from the network. */
void kill_user (const char *source, const char *user, const char *reason)
{
    char *av[2];
    char buf[BUFSIZE];

    if (!user || !*user)
	return;

    if (!source || !*source)
	source = me.name;

    if (!reason)
	reason = "Killed";

    snprintf (buf, sizeof (buf), "%s (%s)", source, reason);

    av[0] = sstrdup (user);
    av[1] = buf;

    send_cmd (source, "%s %s :%s", me.token ? "." : "KILL", user, av[1]);
    do_kill (source, 2, av);
    free (av[0]);
}

#ifdef BAHAMUT
/* Handle an SJOIN command. - Borrowed from IRCServices
	Bahamut:
	    av[0] = TS3 timestamp
	    av[1] = TS3 timestamp - channel creation time
	    av[2] = channel
	    av[3] = channel modes
	    av[4] = limit / key (depends on modes in av[3])
	    av[5] = limit / key (depends on modes in av[3])
	    av[4|5|6] = nickname(s), with modes, joining channel
 */
void do_sjoin (const char *source, int ac, char **av)
{
    User *user;
    char *nick, *t;
    char *channel = av[2];
    char **av_cmode;
    struct u_chanlist *c;
    int modecnt;

    t = av[ac-1];

    while (*(nick = t))
    {
	char modebuf[4] = { 0 };

	t = nick + strcspn (nick, " ");

	if (*t)
	    *t++ = 0;

	modecnt = 0;
	modebuf[0] = '+';

	while (*nick == '@' || *nick == '+')
	{
	    switch (*nick)
	    {
		case '@':
		    modecnt++;
		    modebuf[modecnt] = 'o';
		    break;
		case '+':
		    modecnt++;
		    modebuf[modecnt] = 'v';
		    break;
	    }
	    nick++;
	}

	user = finduser (nick);

	if (!user)
	{
	    if (debuglevel > 1)
		debug ("do_sjoin(): Got SJOIN to %s for nonexistant user %s",
		    channel, nick);
	    continue;
	}

	if (debuglevel > 1)
	    debug ("do_sjoin(): %s joined %s", nick, channel);

	chan_adduser (user, channel);
	c = smalloc (sizeof (*c));
	c->next = user->chans;
	c->prev = NULL;
	if (user->chans)
	    user->chans = c;
	user->chans = c;
	c->chan = findchan (channel);

	if (modecnt > 0)
	{
	    av_cmode = smalloc (sizeof (char *) * (modecnt + 2));
	    av_cmode[0] = channel;
	    av_cmode[1] = modebuf;
	    av_cmode[2] = nick;
	    if (modecnt == 2)
		av_cmode[3] = nick;
	    do_cmode (source, modecnt + 2, av_cmode);
	    free (av_cmode);
	}
    }

    /* Did anyone actually join the channel and are there really any modes? */
    if (av[3] + 1 != '\0')
    {
	/* Set channel modes. */
	av_cmode = smalloc (sizeof (char *) * (ac - 3));
	av_cmode[0] = channel;
	av_cmode[1] = av[3];
	do_cmode (source, ac - 3, av_cmode);
	free (av_cmode);
    }
}
#endif /* BAHAMUT */

